import time
from functools import partial
from typing import Any, Callable, Dict, Iterable, Literal, Tuple, Union

import jax

jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
import jax.tree_util
import optax
from equinox import apply_updates as eqx_apply_updates
from equinox import filter as eqx_filter
from equinox import filter_jit
from equinox import is_inexact_array as eqx_is_inexact_array
from jax import Array as jax_Array
from jax import device_put as jax_device_put
from jax.tree_util import PyTreeDef

from .data_loading import (Path, check_batch_size_validity,
                           load_encoded_training_validation_and_testing_data)
from .equinox_nn_factories import EquinoxMLPWrapper, eqxModule
from .losses import (compute_H1_bochner_relative_errors,
                     compute_L2_bochner_relative_errors,
                     cpu_compute_H1_bochner_relative_errors,
                     cpu_compute_L2_bochner_relative_errors,
                     vectorized_grad_H1_Bochner_loss,
                     vectorized_grad_L2_Bochner_loss)


def __add_norm_keys(data_keys: Tuple[str, ...]) -> Tuple[str, ...]:
    return data_keys + tuple(f"{data_key}_norms" for data_key in data_keys if data_key != "X")


def __create_permuter(N: int, cp_random_seed: int = None) -> Callable:
    # def __create_permuter(N: int, num_arrays: int) -> Callable:

    """
    Creates a callable to permute JAX arrays using permutation indices from CuPy.

    Parameters:
        N (int): Size for generating permutation indices.
        cp_random_seed (int, optional): Seed for reproducible permutation.

    Returns:
        Callable: A function that, given JAX arrays, returns their permuted versions.

    Example:
        >>> permute = create_array_permuter_flat(100, cp_random_seed=42)
        >>> X_permuted, Y_permuted = permute(X, Y)
    """
    import cupy as cp
    from jax.dlpack import from_dlpack as dlpack2jax
    from jax.dlpack import to_dlpack as jax2dlpack

    indices = cp.arange(N)
    if cp_random_seed is not None:
        cp.random.seed(cp_random_seed)

    def permute_arrays(*arrays: Iterable[jax_Array]) -> Tuple[jax_Array, ...]:
        perm = cp.random.permutation(indices)
        # Use DLPack to transfer data between CuPy and JAX and apply permutation
        return tuple(dlpack2jax(cp.from_dlpack(jax2dlpack(arr))[perm]) for arr in arrays)

    return permute_arrays
    # indices = jnp.arange(N)
    # # @jax.jit
    # # def permute_arrays(key, *arrays: Iterable[jax_Array]) -> Tuple[jax_Array, ...]:
    # #     perm = jax.random.permutation(key, indices)
    # #     return tuple(arr[perm] for arr in arrays)

    # if num_arrays == 3:
    #     # @partial(jax.jit, donate_argnums=(1,2,3))
    #     def permute_arrays(key, a, b, c) -> Tuple[jax_Array, jax_Array, jax_Array]:
    #         perm = jax.random.permutation(key, indices)
    #         return a[perm], b[perm], c[perm]
    # elif num_arrays == 2:
    #     @partial(jax.jit, donate_argnums=(1,2))
    #     def permute_arrays(key, a, b) -> Tuple[jax_Array, jax_Array]:
    #         perm = jax.random.permutation(key, indices)
    #         return a[perm], b[perm]

    # return permute_arrays


def __create_slicer(*, batch_size: int, num_input_outputs: int) -> Callable:
    """
    Returns a slicing function that extracts a batch from each array using a fixed batch_size and start index.
    """
    from jax.lax import dynamic_slice_in_dim as jittable_slice

    if num_input_outputs == 2:

        @filter_jit
        def slice_data(X: jax_Array, out_1: jax_Array, end_idx: int) -> Tuple[int, jax_Array, jax_Array, jax_Array]:
            """
            Slices X and out_1 from end_idx for batch_size items.

            Returns:
                Updated end index and the slices of X and out_1.
            """
            return (
                end_idx + batch_size,
                jittable_slice(X, end_idx, batch_size),
                jittable_slice(out_1, end_idx, batch_size),
            )

    elif num_input_outputs == 3:

        @filter_jit
        def slice_data(
            X: jax_Array, out_1: jax_Array, out_2: jax_Array, end_idx: int
        ) -> Tuple[int, jax_Array, jax_Array, jax_Array]:
            """
            Slices X, out_1, and out_2 from end_idx for batch_size items.

            Returns:
            Updated end index and the slices of X, out_1, and out_2.
            """
            return (
                end_idx + batch_size,
                jittable_slice(X, end_idx, batch_size),
                jittable_slice(out_1, end_idx, batch_size),
                jittable_slice(out_2, end_idx, batch_size),
            )

    return slice_data


def __create_optax_optimization_stepper(
    *,
    LOSS_NAME: Literal["L2", "H1"],  # "score", "hybrid", "h1hybrid", "meanVar", "meanEnt", "fisher",
    nn: eqxModule,
    OPTAX_OPTIMIZER_NAME: str,
    learning_rate: Union[float, optax.Schedule],
) -> Tuple[Callable, optax.OptState, PyTreeDef, eqxModule]:
    """
    Creates a jitted optimization stepper for a neural network using an Optax optimizer.

    Parameters:
        LOSS_NAME (Literal["L2", "H1"]): Chooses the gradient function.
        nn (eqxModule): Neural network to optimize.
        OPTAX_OPTIMIZER_NAME (str): Name of the Optax optimizer.
        learning_rate (float | optax.Schedule): Learning rate.

    Returns:
        Tuple containing:
            - A jitted step function for updating the network and optimizer state.
            - The initial flattened optimizer state.
            - The flattened neural network.
            - The network's tree definition.
    """
    gradient = {
        "L2": vectorized_grad_L2_Bochner_loss,
        "H1": vectorized_grad_H1_Bochner_loss,
    }.get(LOSS_NAME)
    if gradient is None:
        raise Exception(f"LOSS_NAME={LOSS_NAME} is not currently implemented")

    optimizer = optax.__getattribute__(OPTAX_OPTIMIZER_NAME)(
        learning_rate=learning_rate,
    )
    optimizer_state = optimizer.init(eqx_filter(nn, eqx_is_inexact_array))

    flat_nn, treedef_nn = jax.tree_util.tree_flatten(nn)
    flat_optimizer_state, treedef_optimizer_state = jax.tree_util.tree_flatten(optimizer_state)
    if LOSS_NAME == "L2":

        @filter_jit
        def take_step(
            flat_optimizer_state: optax.OptState,
            flat_nn: eqxModule,
            X: jax_Array,
            Y: jax_Array,
        ) -> Tuple[optax.OptState, eqxModule]:
            nn = jax.tree_util.tree_unflatten(treedef_nn, flat_nn)
            updates, new_optimizer_state = optimizer.update(
                gradient(nn, X, Y),
                jax.tree_util.tree_unflatten(treedef_optimizer_state, flat_optimizer_state),
                nn,
            )
            new_flat_nn = jax.tree_util.tree_leaves(eqx_apply_updates(nn, updates))
            new_flat_optimizer_state = jax.tree_util.tree_leaves(new_optimizer_state)
            return new_flat_optimizer_state, new_flat_nn

    elif LOSS_NAME == "H1":

        @filter_jit
        def take_step(
            flat_optimizer_state: optax.OptState,
            flat_nn: eqxModule,
            X: jax_Array,
            Y: jax_Array,
            dYdX: jax_Array,
        ) -> Tuple[optax.OptState, eqxModule]:
            nn = jax.tree_util.tree_unflatten(treedef_nn, flat_nn)
            updates, new_optimizer_state = optimizer.update(
                gradient(nn, X, Y, dYdX),
                jax.tree_util.tree_unflatten(treedef_optimizer_state, flat_optimizer_state),
                nn,
            )
            new_flat_nn = jax.tree_util.tree_leaves(eqx_apply_updates(nn, updates))
            new_flat_optimizer_state = jax.tree_util.tree_leaves(new_optimizer_state)
            return new_flat_optimizer_state, new_flat_nn

    return (take_step, flat_optimizer_state, flat_nn, treedef_nn)


def __create_epochs_fn(
    *,
    stepper: Callable[..., Tuple[optax.OptState, eqxModule]],
    slicer: Callable,
    random_permuter: Callable[..., Tuple[jnp.ndarray, ...]],
    batches_arange: jnp.ndarray,
    epochs_arange: jnp.ndarray,
    cpu_to_gpu_n_batches: int = 1,
) -> Tuple[optax.OptState, eqxModule]:
    """
    Executes a training loop over epochs using batched and permuted training data.

    Parameters:
        stepper: Function that updates the optimizer state and neural network using a batch.
        slicer: Function that slices data arrays into batches.
        nn: The neural network model.
        optimizer_state (optax.OptState): Current state of the optimizer.
        random_permuter: Function that randomly permutes the training data arrays.
        batches_arange: Array of epoch indices for iterating over the data.
        training_data: Tuple of training arrays (e.g., features, labels).

    Returns:
        Tuple[optax.OptState, eqxModule]: The final optimizer state and the trained model.
    """
    import equinox as eqx

    if cpu_to_gpu_n_batches == 1:

        def epochs_fn(optimizer_state, nn, training_data):
            for _ in epochs_arange:
                training_data = random_permuter(*training_data)

                end_idx = 0

                for j in batches_arange:
                    end_idx, *batch = slicer(*training_data, end_idx)  # X, *other_data
                    optimizer_state, nn = stepper(optimizer_state, nn, *batch)
            return optimizer_state, nn  # , X, fX, dfXdX

    else:
        cpu_to_gpu_n_batches_arange = jnp.arange(cpu_to_gpu_n_batches)

        def epochs_fn(optimizer_state, nn, training_data):
            for _ in epochs_arange:
                X, *other_data = random_permuter(*training_data)  # permuteon cpu
                for k in cpu_to_gpu_n_batches_arange:
                    # X_gpu,* other_data_gpu =  transfer part of X to gpu

                    end_idx = 0
                    for j in batches_arange:
                        end_idx, X_batch, *other_batch = slicer(X_gpu, *other_data_gpu, end_idx)
                        optimizer_state, nn = stepper(optimizer_state, nn, X_batch, *other_batch)
            return optimizer_state, nn

    return epochs_fn


def load_encoded_data_train_and_test_dino(
    data_path: Path,
    eqx_wrapper: EquinoxMLPWrapper,
    N_MAX_EPOCHS: int,
    STEP_SIZE: float,
    BATCH_SIZE: int,
    OPTAX_OPTIMIZER_NAME: str,
    LOSS_NAME: str,
    N_TRAIN: int,
    REDUCED_DIMS: Tuple[int, int],
    RANDOM_PERMUTATIONS_SEED: int,
    N_VAL: int = 2500,
) -> Tuple[EquinoxMLPWrapper, Dict[str, Any]]:
    import ast
    from pathlib import Path

    training_dir = Path(data_path, "training")

    # grab one npy file to inspect tuple
    example_file = next(training_dir.glob("*.npy"))

    tuple_str = example_file.name.split(")")[0] + ")"
    input_reduced_dim, output_reduced_dim = ast.literal_eval(tuple_str)

    input_key = "encoded_inputs" if input_reduced_dim is not None else "inputs"
    output_key = "encoded_outputs" if output_reduced_dim is not None else "outputs"

    if LOSS_NAME == "L2":
        training_data_keys = (input_key, output_key)
    elif LOSS_NAME == "H1":
        training_data_keys = (input_key, output_key, "encoded_Jacobians")
    else:
        raise ValueError(f"Unknown LOSS_NAME: {LOSS_NAME}")
    if training_data_keys is None:
        raise Exception("Not currently implemented")
    renaming_dict = {
        "encoded_inputs": "X",
        "encoded_outputs": "fX",
        "inputs": "X",
        "outputs":"fX",
        "encoded_Jacobians": "dfXdX",
    }
    train_val_test = load_encoded_training_validation_and_testing_data(
        data_path=data_path,
        REDUCED_DIMS=REDUCED_DIMS,
        training_data_keys=training_data_keys,
        renaming_dict=renaming_dict,
        N_TRAIN=N_TRAIN,
        N_VAL=N_VAL,
    )
    nn, results = train_and_test_dino(
        nn=eqx_wrapper.nn,
        N_MAX_EPOCHS=N_MAX_EPOCHS,
        STEP_SIZE=STEP_SIZE,
        BATCH_SIZE=BATCH_SIZE,
        train_val_test=train_val_test,
        OPTAX_OPTIMIZER_NAME=OPTAX_OPTIMIZER_NAME,
        RANDOM_PERMUTATIONS_SEED=RANDOM_PERMUTATIONS_SEED,
        training_data_keys=("X", "fX", "dfXdX") if LOSS_NAME == "H1" else ("X", "fX"),
        LOSS_NAME=LOSS_NAME,
    )
    eqx_wrapper.params = nn  # is this right?
    return eqx_wrapper, results


def train_and_test_dino(
    nn: eqxModule,
    N_MAX_EPOCHS: int,
    STEP_SIZE: float,
    BATCH_SIZE: int,
    train_val_test: dict,  # test has to have Jacobians, 'dfXdX' in them!!!
    N_EPOCHS_BETWEEN_TEST: int = 100,
    OPTAX_OPTIMIZER_NAME: str = "adam",
    RANDOM_PERMUTATIONS_SEED: int = 0,
    training_data_keys: Tuple[str, ...] = ("X", "fX", "dfXdX"),
    testing_data_keys: Tuple[str, ...] = ("X", "fX", "dfXdX"),
    LOSS_NAME: str = "H1",
) -> Tuple[eqxModule, Dict[str, Any]]:
    # Place on GPU!

    if not all(list(arr.devices())[0].platform == "gpu" for arr in train_val_test["train"].values()):
        print("Training data did not yet reside on GPU. Placing on GPU.")
        training_data = tuple(jax_device_put(train_val_test["train"][k]) for k in training_data_keys)
    else:
        training_data = tuple(train_val_test["train"][k] for k in training_data_keys)

    cpu_test_data = train_val_test["test"]
    val_data = train_val_test.get("val")
    if val_data is not None:
        assert all(
            list(arr.devices())[0].platform == "gpu" for arr in val_data.values()
        ), "All arrays in train_val_test['val'] should be on the GPU!"
    assert all(
        list(arr.devices())[0].platform != "gpu" for arr in cpu_test_data.values()
    ), "All arrays in train_val_test['test'] should be on the CPU!"

    # Prepare for training
    N_TRAIN = training_data[0].shape[0]
    n_train_batches = check_batch_size_validity(data_iterable=training_data, batch_size=BATCH_SIZE)
    N_MAX_OUTER_ITERS = N_MAX_EPOCHS // N_EPOCHS_BETWEEN_TEST
    num_train_steps = N_MAX_EPOCHS * n_train_batches
    lr_schedule = optax.linear_schedule(
        init_value=STEP_SIZE, transition_steps=num_train_steps, end_value=STEP_SIZE * 0.3
    )
    slicer = __create_slicer(batch_size=BATCH_SIZE, num_input_outputs=len(training_data))
    random_permuter = __create_permuter(
        N_TRAIN, RANDOM_PERMUTATIONS_SEED
    )  # len(training_data_keys)) #, RANDOM_PERMUTATIONS_SEED)
    stepper, optimizer_state, nn, nn_treedef = __create_optax_optimization_stepper(
        LOSS_NAME=LOSS_NAME,
        nn=nn,
        OPTAX_OPTIMIZER_NAME=OPTAX_OPTIMIZER_NAME,
        learning_rate=lr_schedule,
    )
    n_epochs = __create_epochs_fn(
        slicer=slicer,
        random_permuter=random_permuter,
        stepper=stepper,
        batches_arange=jnp.arange(n_train_batches),
        epochs_arange=jnp.arange(N_EPOCHS_BETWEEN_TEST),
    )

    compute_relative_errs = {  # harded coded
        "L2": compute_L2_bochner_relative_errors,
        "H1": compute_H1_bochner_relative_errors,
    }.get(LOSS_NAME)

    cpu_compute_relative_errs = {  # harded coded
        "L2": cpu_compute_L2_bochner_relative_errors,
        "H1": cpu_compute_H1_bochner_relative_errors,
    }.get(LOSS_NAME)
    print("Started training...")

    total_epoch_time = 0.0
    total_validation_time = 0.0
    validations = dict()
    results = dict()
    val_data_rel_errors_keys = __add_norm_keys(training_data_keys)
    test_data_rel_errors_keys = __add_norm_keys(testing_data_keys)

    for outer_iter in jnp.arange(N_MAX_OUTER_ITERS):
        start = time.perf_counter()
        optimizer_state, nn = n_epochs(optimizer_state, nn, training_data)
        total_epoch_time += time.perf_counter() - start

        # evaluate validation error
        if val_data:
            print(f"Iter: {(outer_iter+1)*N_EPOCHS_BETWEEN_TEST},  computing validation errors")
            start = time.perf_counter()
            nn_host = jax.device_get(nn)
            nn_pytree_host = jax.tree_util.tree_unflatten(nn_treedef, nn_host)
            nn_pytree = jax.tree_util.tree_unflatten(nn_treedef, nn)
            train_errors = cpu_compute_relative_errs(
                nn_pytree_host,
                *(train_val_test["train_cpu"][key] for key in val_data_rel_errors_keys),
            )
            validation_errors = compute_relative_errs(
                nn_pytree,
                *(val_data[key] for key in val_data_rel_errors_keys),
            )
            total_validation_time += time.perf_counter() - start
            print("val errors", validation_errors)
            print("train_errors", train_errors)
            # break if validation_loss is not decreasing
    validations["total_validation_time"] = total_validation_time
    # Testing error
    nn_host = jax.device_get(nn)

    nn_pytree = jax.tree_util.tree_unflatten(nn_treedef, nn)
    nn_pytree_host = jax.tree_util.tree_unflatten(nn_treedef, nn_host)
    results["test_errors"] = cpu_compute_H1_bochner_relative_errors(
        nn_pytree_host,
        *(cpu_test_data[key] for key in test_data_rel_errors_keys),
    )
    results["total_training_time"] = total_epoch_time
    results["total_training_time_minus_validation"] = total_epoch_time - total_validation_time
    return nn_pytree, results



import json
import math
from pathlib import Path
from typing import Any, Dict, List

import equinox as eqx
import jax

jax.config.update("jax_enable_x64", True)
import jax.random as jr
from equinox import Module as eqxModule
from pydantic import (BaseModel, ConfigDict, field_serializer, field_validator,
                      model_serializer, model_validator)


# equinox_nn_factories
class EquinoxMLPConfig(BaseModel, validate_assignment=True):
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=False)

    input_dimension: int
    output_dimension: int
    random_initializer_key: Any
    activation_name: str
    layer_width: int
    depth: int

    @field_serializer("random_initializer_key")
    def serialize_random_initializer_key(self, key: jax.Array, _info) -> list:
        """Convert JAX PRNGKey to a list."""
        # Note: adjust this conversion if needed
        return jax.random.key_data(key).tolist()[1]

    @field_validator("random_initializer_key", mode="before")
    @classmethod
    def create_random_initializer_key(cls, key_int: int) -> jax.Array:
        """Reconstruct a PRNGKey from an integer."""
        return jr.PRNGKey(key_int)


def build_equinox_MLP_from_config(eqx_config: EquinoxMLPConfig) -> eqxModule:
    """
    Build an Equinox MLP from the provided configuration.
    """
    from equinox.nn import MLP

    return MLP(
        in_size=eqx_config.input_dimension,
        out_size=eqx_config.output_dimension,
        width_size=eqx_config.layer_width,
        depth=eqx_config.depth,
        activation=eqx.filter_jit(getattr(jax.nn, "gelu")),
        key=eqx_config.random_initializer_key,
    )


def build_equinox_MLP_from_config_and_load_weights(*, eqx_path: Path, eqx_config: EquinoxMLPConfig = None) -> eqxModule:
    """
    Build an Equinox MLP from the configuration (or load it from disk if not provided)
    and then load its weights from file.
    """
    if eqx_config is None:
        with open(Path(eqx_path, "config.json"), "r") as f:
            config_data = json.load(f)
        eqx_config = EquinoxMLPConfig(**config_data)
    mlp = build_equinox_MLP_from_config(eqx_config)
    return eqx.tree_deserialise_leaves(Path(eqx_path, "weights.eqx"), mlp)


def partion_eqx_nn_by_trainability(eqx_nn: eqxModule):
    """
    Partition eqx Module parameters into trainable and non-trainable groups.
    """
    return eqx.partition(eqx_nn, eqx.is_array)


def _truncated_normal_initializer(weight: jax.Array, key: jr.PRNGKey) -> jax.Array:
    """
    Apply a truncated normal initialization to a weight array.

    This initializer computes a standard deviation based on the input dimension,
    and returns a new weight array with values drawn from a truncated normal distribution.
    """
    out_dim, in_dim = weight.shape
    stddev = math.sqrt(1 / in_dim)
    return stddev * jax.random.truncated_normal(key, shape=(out_dim, in_dim), lower=-2, upper=2)


def _get_nn_weights(model: eqxModule) -> List[jax.Array]:
    """
    Retrieve the weight matrices from all linear layers in the model.
    """
    is_linear = lambda x: isinstance(x, eqx.nn.Linear)
    return [x.weight for x in jax.tree_util.tree_leaves(model, is_leaf=is_linear) if is_linear(x)]


def _init_linear_layer_weights(model: eqxModule, key: jr.PRNGKey) -> eqxModule:
    """
    Reinitialize all linear layers in the model with a truncated normal initializer.

    Use this function if you wish to override the default initialization.
    """
    weights = _get_nn_weights(model)
    subkeys = jr.split(key, len(weights))
    new_weights = [_truncated_normal_initializer(w, sk) for w, sk in zip(weights, subkeys)]
    new_model = eqx.tree_at(_get_nn_weights, model, new_weights)
    return new_model


class EquinoxMLPWrapper(BaseModel, validate_assignment=False):
    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=False)

    params: Any
    static: Any
    eqx_config: EquinoxMLPConfig
    path: Path

    @property
    def nn(self):
        return eqx.combine(self.params, self.static)

    @model_serializer
    def serialize_everything(self, _info) -> Dict[str, Any]:
        self.path.mkdir(parents=True, exist_ok=True)
        print(f"saving eqx weights and config to {self.path}")
        eqx.tree_serialise_leaves(Path(self.path, "weights.eqx"), self.nn)
        with open(Path(self.path, "config.json"), "w", encoding="utf-8") as f:
            json.dump(self.eqx_config.model_dump(), f, indent=4)
        return {"path": self.path, "eqx_config": self.eqx_config.model_dump()}

    @model_validator(mode="before")
    @classmethod
    def create_model(cls, values):
        path = values.get("path")
        if not path:
            raise ValueError("Missing required key: path")
        eqx_config = values.get("eqx_config")
        if not isinstance(eqx_config, EquinoxMLPConfig):
            # If eqx_config is not a dict, load it from file if load_from_file is specified.
            if not isinstance(eqx_config, dict) and values.get("load_from_file"):
                print("Loading config from file")
                with open(Path(path, "config.json")) as json_file:
                    eqx_config = json.load(json_file)
            else:
                print("Not loading config from file--using passed in eqx_config dict")
            eqx_config = EquinoxMLPConfig(**eqx_config)
        values["eqx_config"] = eqx_config

        if values.get("load_from_file"):
            print("Loading eqx MLP from file")
            eqx_nn = build_equinox_MLP_from_config_and_load_weights(eqx_config=eqx_config, eqx_path=values["path"])
        else:
            print("Building new eqx MLP from scratch")
            eqx_nn = build_equinox_MLP_from_config(eqx_config=eqx_config)
            eqx_nn = _init_linear_layer_weights(eqx_nn, eqx_config.random_initializer_key)
        values["params"], values["static"] = partion_eqx_nn_by_trainability(eqx_nn)
        return values


from typing import Union


def config_and_eqx_module(
    *,
    eqx_config: EquinoxMLPConfig,
    eqx_nn: Union[eqx.Module, None] = None,
    eqx_path: Union[Path, None] = None,
):
    if not isinstance(eqx_config, EquinoxMLPConfig):
        eqx_config = EquinoxMLPConfig(**eqx_config.copy())
    if not isinstance(eqx_nn, eqx.Module):
        assert isinstance(eqx_path, Path)
        eqx_nn = build_equinox_MLP_from_config_and_load_weights(eqx_config=eqx_config, eqx_path=eqx_path)
    return eqx_config, eqx_nn



import gc
import time
from operator import itemgetter
from pathlib import Path
from typing import Iterable, List, Literal, Tuple, Union

import numpy as np
from bayesflux import (hippylib_sampler_from_model,
                       multiprocess_generate_hippylib_full_Jacobian_data,
                       multiprocess_generate_hippylib_output_data,
                       multiprocess_generate_reduced_hippylib_training_data)
from bayesflux.generation import (GaussianInputOuputAndDerivativesSampler,
                                  generate_full_Jacobian_data,
                                  generate_output_data,
                                  generate_reduced_training_data)
from jax.typing import ArrayLike as JAXArrayLike
from numpy.typing import ArrayLike

from .data_loading import (device_get_pytree, dump_pytree_to_disk,
                           load_pytree_from_disk)


def __find_eigs_truncation_idxs(
    lambda_seq: List[float], mu_seq: List[float], percentage: float, threshold: float = 100  # .0
) -> Tuple[int, int]:
    """
    Find indices at which to truncate ordered eigenvalue sequences to capture a given
    percentage of the tail sum of each eigenvalue sequences below a specified threshold.

    Parameters:
        lambda_seq (List[float]):
            A list of eigenvalues (sorted in descending order).

        mu_seq (List[float]):
            Another list of eigenvalues (sorted in descending order).

        percentage (float):
            % of the sum of eigenvalues below the threshold we wish to capture.

        threshold (float, optional):
            The eigenvalue threshold defining the "tail" region (default is 1).
            Choose np.inf for all eigenvalues.

    Returns:
        Tuple[int, int]:
            A tuple containing the indices (r_lambda, r_mu) indicating how many
            eigenvalues below the threshold in each sequence should be included
            to reach at least the specified percentage of the tail sum.

    Notes:
        - The input sequences must be sorted in descending order.
        - The returned indices are counts of eigenvalues below the threshold included
          from the start of the tail region.
    """
    # Filter eigenvalues below the threshold
    lambda_tail = [val for val in lambda_seq if val < threshold]
    mu_tail = [val for val in mu_seq if val < threshold]
    rest_lambda = len(lambda_seq) - len(lambda_tail)
    rest_mu = len(mu_seq) - len(mu_tail)

    # Calculate the total sum of eigenvalues in the tails
    total_tail_sum = sum(lambda_tail) + sum(mu_tail)
    target_sum = percentage * total_tail_sum

    # Initialize indices and running sum
    i, j = 0, 0
    curr_sum = 0.0

    # Accumulate eigenvalues from tails until the target sum is reached
    while curr_sum < target_sum:
        # Next available eigenvalue from lambda_tail or negative infinity if exhausted
        next_lambda = lambda_tail[i] if i < len(lambda_tail) else -float("inf")
        next_mu = mu_tail[j] if j < len(mu_tail) else -float("inf")

        # Select the larger available eigenvalue to accumulate more quickly
        if next_lambda >= next_mu:
            curr_sum += next_lambda
            i += 1
        else:
            curr_sum += next_mu
            j += 1
    print("lambdas:", lambda_seq)
    print("mu:", mu_seq)

    return i + rest_lambda, j + rest_mu


def generate_moment_based_encoder_decoders(
    key,
    input_covariance_matrix=None,
    L2_inner_product_matrix=None,
    output_samples: ArrayLike = None,
    input_output_dim_percentage_pairs: Iterable[float] = None,
    input_output_dim_pairs: Iterable[Tuple[int, int]] = None,
    input_dims: Iterable[int] = None,
    output_dims: Iterable[int] = None,  # only one of input_dims / output_dims should be provided
):
    import jax

    if input_output_dim_percentage_pairs:
        max_input_dimension = input_covariance_matrix.shape[0]
        max_output_dimension = output_samples.shape[1]
    elif input_output_dim_pairs:
        max_input_dimension = max(input_output_dim_pairs, key=itemgetter(0))[0]
        max_output_dimension = max(input_output_dim_pairs, key=itemgetter(1))[1]
    elif input_dims:
        max_input_dimension = max(input_dims)
        max_output_dimension = None
        input_output_dim_pairs = [(input_dim, None) for input_dim in input_dims]
    elif output_dims:
        max_input_dimension = None
        max_output_dimension = max(output_dims)
        input_output_dim_pairs = [(None, output_dim) for output_dim in output_dims]
    else:
        raise Exception("No reduction dimensions provided!")

    from bayesflux.subspace_detection import moment_based_dimension_reduction

    # Compute the input/output encodec given maximum truncation dimensions
    full_encodecs = moment_based_dimension_reduction(
        key=key,
        output_samples=jax.device_put(output_samples) if output_samples is not None else output_samples,
        input_covariance_matrix=input_covariance_matrix,
        L2_inner_product_matrix=L2_inner_product_matrix,
        max_input_dimension=max_input_dimension,
        max_output_dimension=max_output_dimension,
    )

    if input_output_dim_percentage_pairs:
        input_output_dim_pairs = [
            __find_eigs_truncation_idxs(
                full_encodecs["input"]["eigenvalues"], full_encodecs["output"]["eigenvalues"], percentage
            )
            for percentage in input_output_dim_percentage_pairs
        ]
    encodecs = dict()
    for input_dim_r, output_dim_r in input_output_dim_pairs:
        encodecs[(input_dim_r, output_dim_r)] = dict()
        for k1, dim_r in zip(("input", "output"), (input_dim_r, output_dim_r)):
            encodecs[(input_dim_r, output_dim_r)][k1] = dict()
            if dim_r is not None:
                for k2 in ("eigenvalues", "encoder", "decoder"):
                    encodecs[(input_dim_r, output_dim_r)][k1][k2] = full_encodecs[k1][k2].copy()[..., :dim_r]
    return encodecs


def generate_derivative_informed_encoder_decoders_from_full_jacobians(
    key,
    prior_precision: ArrayLike,
    noise_precision: float,
    full_jacobian_data: ArrayLike,
    input_output_dim_percentage_pairs: Iterable[float] = None,
    input_output_dim_pairs: Iterable[Tuple[int, int]] = None,
    input_dims: Iterable[int] = None,
    output_dims: Iterable[int] = None,
):
    import jax
    import jax.numpy as jnp

    if input_output_dim_percentage_pairs:
        max_input_dimension = full_jacobian_data.shape[2]
        max_output_dimension = full_jacobian_data.shape[1]
        print("max input dim", max_input_dimension)
        print("max output dim", max_output_dimension)
    elif input_output_dim_pairs:
        max_input_dimension = max(input_output_dim_pairs, key=itemgetter(0))[0]
        max_output_dimension = max(input_output_dim_pairs, key=itemgetter(1))[1]
    elif input_dims:
        max_input_dimension = max(input_dims)
        max_output_dimension = None
        input_output_dim_pairs = [(input_dim, None) for input_dim in input_dims]
    elif output_dims:
        max_input_dimension = None
        max_output_dimension = max(output_dims)
        input_output_dim_pairs = [(None, output_dim) for output_dim in output_dims]
    else:
        raise Exception("No reduction dimensions provided!")
    from bayesflux.subspace_detection import \
        information_theoretic_dimension_reduction

    # Compute the input/output encodec given maximum truncation dimensions
    full_encodecs = information_theoretic_dimension_reduction(
        key=key,
        J_samples=jax.device_put(full_jacobian_data),
        noise_precision=jax.device_put(noise_precision),
        prior_precision=jax.device_put(prior_precision),
        prior_covariance=jnp.linalg.inv(prior_precision),
        max_input_dimension=max_input_dimension,
        max_output_dimension=max_output_dimension,
    )
    if input_output_dim_percentage_pairs:
        input_output_dim_pairs = [
            __find_eigs_truncation_idxs(
                full_encodecs["input"]["eigenvalues"], full_encodecs["output"]["eigenvalues"], percentage
            )
            for percentage in input_output_dim_percentage_pairs
        ]
        print("\t\tTotal input eigenvalues:", np.sum(full_encodecs["input"]["eigenvalues"]))
        print("\t\tTotal output eigenvalues:", np.sum(full_encodecs["output"]["eigenvalues"]))

        print("\t\tPercentages:", input_output_dim_percentage_pairs)
        print("\t\tDims:", input_output_dim_pairs)

        #TODO remove eventually
        # hp_eigs_input, hp_input_encoder, hp_input_decoder = hippylib_active_subspace(full_jacobian_data, noise_variance, prior, 2145, 10)

        # eigs_input = encoder_decoders_dict['input']['eigenvalues']
        # eigs_output = encoder_decoders_dict['output']['eigenvalues']
        # plt.semilogy(eigs_input[jnp.argsort(eigs_input)[::-1]])
        # plt.semilogy(hp_eigs_input[jnp.argsort(hp_eigs_input)[::-1]])
        # # plt.semilogy(eigs_input[jnp.argsort(eigs_input)[::-1]],label="estimated")

        # # plt.semilogy(hp_eigs_input[jnp.argsort(hp_eigs_input)[::-1]],label="true_hp")
        # plt.title("Eigenvalues of C E[J^T Gamma^{-1}_N J]")
        # # plt.legend()
        # plt.show()

        # plt.semilogy(eigs_output[jnp.argsort(eigs_output)[::-1]])
        # # plt.semilogy(eigs_input[jnp.argsort(eigs_input)[::-1]],label="estimated")

        # # plt.semilogy(hp_eigs_input[jnp.argsort(hp_eigs_input)[::-1]],label="true_hp")
        # plt.title("Eigenvalues of Gamma^{-1}_N E[J^T C_N J]")
        # # plt.legend()
        # plt.show()

        # # print(f"Use only {input_reduced_dim} eigenfunctions, percent of spectrum", 100*sum(eigs_input)/sum(true_eigs_input))
    encodecs = dict()
    for input_dim_r, output_dim_r in input_output_dim_pairs:
        encodecs[(input_dim_r, output_dim_r)] = dict()
        # TODO: remove eventually
        # print(f"Use only {input_reduced_dim} eigenfunctions, percent of TRUE input spectrum", 100*sum(hp_eigs_input[0:input_reduced_dim])/sum(hp_eigs_input))
        # # print(f"Use only {input_reduced_dim} eigenfunctions, percent of TRUE input+output spectrum", 100*sum(hp_eigs_input[0:input_reduced_dim])/(sum(true_eigs_output)+sum(hp_eigs_input)))

        # print("Trace of expected J^TGamma^-1 J", sum(hp_eigs_input))
        # print("Trace of intput tail sum", sum(hp_eigs_input[input_reduced_dim:]))

        # # print(f"Use only {output_reduced_dim} eigenfunctions, percent of spectrum", 100*sum(eigs_output)/sum(true_eigs_output)) #1,2 5, 50
        # print(f"Use only {output_reduced_dim} eigenfunctions, percent of TRUE output spectrum", 100*sum(eigs_output[0:output_reduced_dim])/sum(eigs_output)) #1,2 5, 50
        # print(f"Use only {input_reduced_dim, output_reduced_dim} eigenfunctions, percent of TRUE intput+output spectrum", (100*sum(eigs_output[0:output_reduced_dim])+ 100*sum(hp_eigs_input[0:input_reduced_dim]))/(sum(eigs_output)+sum(hp_eigs_input))) #1,2 5, 50

        # print("Trace of expected JCJ^T", sum(eigs_output))
        # print("Trace of output tail sum", sum(eigs_output[output_reduced_dim:]),"\n\n")
        # print("reduced dimensions", input_dim_r, output_dim_r)
        for k1, dim_r in zip(("input", "output"), (input_dim_r, output_dim_r)):
            encodecs[(input_dim_r, output_dim_r)][k1] = dict()
            if dim_r is not None:
                for k2 in ("eigenvalues", "encoder", "decoder"):
                    encodecs[(input_dim_r, output_dim_r)][k1][k2] = full_encodecs[k1][k2].copy()[..., :dim_r]
    del full_encodecs
    return encodecs  # needs to be saved to model's name


def generate_latent_data_for_dino_training(
    N_trains: Iterable[int],
    key: Union[JAXArrayLike, int],
    dimension_reduction_type: Literal["derivative_informed", "moment_based"],
    model_obj: GaussianInputOuputAndDerivativesSampler = None,
    model_name: str = None,
    input_output_dim_percentage_pairs: Iterable[float] = None,
    input_output_dim_pairs: Iterable[Tuple[int, int]] = None,
    input_dims: Iterable[int] = None,
    output_dims: Iterable[int] = None,
    N_encodec_computation: int = 1000,
    N_test: int = 25_000,
    N_val: int = 2500,
    save_path: Path = None,
    reduce_input_before: bool = False,
    N_multiprocessing: int = 1,
):
    N_train_max = max(N_trains)
    N_reduced_samples = N_train_max + N_test + N_val
    subspace_start = time.time()
    if dimension_reduction_type == "derivative_informed":
        print("Generating data for derivative informed subspace detection")
        if N_multiprocessing >= 1:
            subspace_detect_data = multiprocess_generate_hippylib_full_Jacobian_data(
                model_name=model_name,
                N_samples=N_encodec_computation,
                N_processes=N_multiprocessing,
                print_progress=True,
            )
            model_obj = hippylib_sampler_from_model(model_name, 42)
        else:
            subspace_detect_data = generate_full_Jacobian_data(
                sampler_wrapper=model_obj, N_samples=N_encodec_computation
            )
            # print("loaded_results",susbpace_detect_data)
            # input_output_reduced_dim_percentage_pairs  = [0.999, 0.99,0.95,0.9 ] #(2145, 10), (2145,25)
        print("Computing derivative informed encodecs")
        gc.collect()
        encodec_dict = generate_derivative_informed_encoder_decoders_from_full_jacobians(
            key=key,
            prior_precision=model_obj.precision,
            noise_precision=model_obj.noise_precision,
            full_jacobian_data=subspace_detect_data["Jacobians"],
            input_output_dim_percentage_pairs=input_output_dim_percentage_pairs,
            input_output_dim_pairs=input_output_dim_pairs,
            input_dims=input_dims,
            output_dims=output_dims,
        )

    elif dimension_reduction_type == "moment_based":
        print("Generating data for moment-based subspace detection")
        if not input_dims:
            if N_multiprocessing >= 1:
                del model_obj
                gc.collect()
                subspace_detect_data = multiprocess_generate_hippylib_output_data(
                    model_name=model_name,
                    N_samples=N_encodec_computation,
                    N_processes=N_multiprocessing,
                    print_progress=True,
                )
            else:
                subspace_detect_data = generate_output_data(sampler_wrapper=model_obj, N_samples=N_encodec_computation)
        encodec_dict = generate_moment_based_encoder_decoders(
            key=key,
            input_covariance_matrix=np.linalg.inv(model_obj.precision) if output_dims is None else None,
            L2_inner_product_matrix=model_obj.L2_inner_product_matrix if output_dims is None else None,
            output_samples=subspace_detect_data["outputs"] if input_dims is None else None,
            input_output_dim_percentage_pairs=input_output_dim_percentage_pairs,
            input_output_dim_pairs=input_output_dim_pairs,
            input_dims=input_dims,
            output_dims=output_dims,
        )
    gc.collect()

    if save_path:
        if not (input_dims and dimension_reduction_type == "moment_based"):
            print("\t\tDumping this data to disk")
            Path(save_path, dimension_reduction_type).mkdir(parents=True, exist_ok=True)
            path = Path(save_path, dimension_reduction_type, "subpace_detect_data.hkl")
            dump_pytree_to_disk(subspace_detect_data, path)
            # print(f"Checking saved correctly to {path}")
            # subspace_detect_data = load_pytree_from_disk(path)
            for k in subspace_detect_data.keys():
                try:
                    print(k, subspace_detect_data[k].device())
                except:
                    pass

        print("\t\tDumping this data to disk")
        path = Path(save_path, dimension_reduction_type, "encodec_dict.hkl")
        dump_pytree_to_disk(encodec_dict, path)
        # print(f"Checking saved correctly to {path}")
        encodec_dict = load_pytree_from_disk(path)

        if input_output_dim_pairs is None:
            if input_dims:
                input_output_dim_pairs = [(input_dim, None) for input_dim in input_dims]
            elif output_dims:
                input_output_dim_pairs = [(None, output_dim) for output_dim in output_dims]
        for REDUCED_DIMS in input_output_dim_pairs:
            encodec_dict_i = encodec_dict[REDUCED_DIMS]
            import jax.numpy as jnp

            print(
                "\t\t", jnp.linalg.norm(
                    encodec_dict_i["input"]["encoder"].T @ encodec_dict_i["input"]["decoder"]
                    - jnp.eye(encodec_dict_i["input"]["encoder"].shape[1])
                )
                / jnp.linalg.norm(jnp.eye(encodec_dict_i["input"]["encoder"].shape[1]))
            )
            if encodec_dict_i["output"].get("encoder"):
                print(
                    "\t\t", jnp.linalg.norm(
                        encodec_dict_i["output"]["encoder"].T @ encodec_dict_i["output"]["decoder"]
                        - jnp.eye(encodec_dict_i["output"]["encoder"].shape[1])
                    )
                    / jnp.linalg.norm(jnp.eye(encodec_dict_i["output"]["encoder"].shape[1]))
                )

    if not (input_dims and dimension_reduction_type == "moment_based"):
        from bayesflux.encoding import encode_input_output_Jacobian_data

        print("Encoding subspace detection training samples")
        reduced_subspace_detect_data_dicts = dict()
        for reduced_dims_tuple_i, encodec_dict_i in encodec_dict.items():
            print("\t\t...input/output reduced dimensions:", reduced_dims_tuple_i)
            reduced_subspace_detect_data_dicts[reduced_dims_tuple_i] = encode_input_output_Jacobian_data(
                inputs=subspace_detect_data["inputs"],
                outputs=subspace_detect_data["outputs"],
                jacobians=subspace_detect_data.get("Jacobians"),  # If Jacobians are not present, it still works
                input_encoder=encodec_dict_i["input"].get("encoder"),
                output_encoder=encodec_dict_i["output"].get("encoder"),
                input_decoder=encodec_dict_i["input"].get("decoder"),
            )
            print(
                "\t\tEncoding times (input, output, Jacobian): ",
                reduced_subspace_detect_data_dicts[reduced_dims_tuple_i].get(
                    "input_encoding_time", "No input dimension reduction"
                ),
                reduced_subspace_detect_data_dicts[reduced_dims_tuple_i].get(
                    "output_encoding_time", "No output dimension reduction"
                ),
                reduced_subspace_detect_data_dicts[reduced_dims_tuple_i].get(
                    "Jacobian_encoding_time", "No Jacobians to encode"
                ),
            )
            reduced_subspace_detect_data_dicts[reduced_dims_tuple_i]["encodec"] = encodec_dict_i

        if save_path:
            print("\t\tDumping this data to disk")
            path = Path(save_path, dimension_reduction_type, "subspace_detection_data_reduced_and_encoders.hkl")
            dump_pytree_to_disk(
                reduced_subspace_detect_data_dicts, path
            )  # name should reflect apparoach to dim reduction
            # print(f"\tChecking saved correctly to {path}")
            # reduced_subspace_detect_data_dicts = load_pytree_from_disk(path)

    reduced_all_data = dict()
    print("\t\tGetting encoders from gpu, placing back on cpu")
    encodec_dict_on_cpu = device_get_pytree(encodec_dict)
    del encodec_dict, reduced_subspace_detect_data_dicts
    gc.collect()
    subspace_time = time.time() - subspace_start
    print("\t\tTotal subspace computation computation time", subspace_time)
    print(
        "Generating remaining training data in encodec latent space, for each of the N encodecs, the same samples will be generated (with different dimension reductions)."
    )
    for reduced_dims_tuple_i, encodec_dict_i in encodec_dict_on_cpu.items():
        print("\t\t...input/output reduced dimensions:", reduced_dims_tuple_i)
        if N_multiprocessing >= 1:
            print(f"\t\tUsing random seeds offset by {N_multiprocessing} from previous random seed")
            reduced_all_data[reduced_dims_tuple_i] = multiprocess_generate_reduced_hippylib_training_data(
                model_name,
                N_samples=N_reduced_samples,
                N_processes=N_multiprocessing,
                input_encoder=encodec_dict_i["input"].get("encoder"),
                output_encoder=encodec_dict_i["output"].get(
                    "encoder"
                ),  # np.identity(subspace_detect_data["outputs"].shape[1]), #
                input_decoder=encodec_dict_i["input"].get("decoder"),
                random_seed_offset=N_multiprocessing,
                reduce_input_before=reduce_input_before,
            )
        else:
            reduced_all_data[reduced_dims_tuple_i] = generate_reduced_training_data(
                sampler_wrapper=model_obj,
                N_samples=N_reduced_samples,
                input_encoder=encodec_dict_i["input"].get("encoder"),
                output_encoder=encodec_dict_i["output"].get(
                    "encoder"
                ),  # np.identity(subspace_detect_data["outputs"].shape[1]), #
                input_decoder=encodec_dict_i["input"].get("decoder"),
                reduce_input_before=reduce_input_before,
            )
        reduced_all_data[reduced_dims_tuple_i]["encodec"] = encodec_dict_i
        print("data keys", reduced_all_data[reduced_dims_tuple_i].keys())
        if dimension_reduction_type == "derivative_informed":
            print("\t\tjacobian encoding time", reduced_all_data[reduced_dims_tuple_i]["Jacobian_encoding_time"])
            if encodec_dict_i["input"].get("decoder") is not None:
                print("\t\tjacobian decoding time", reduced_all_data[reduced_dims_tuple_i]["Jacobian_decoding_time"])
    if save_path:
        print("\t\tDumping latent train, test, val data to disk; returning test_train_val dictionary.")
    reduced_train_test_val = dict()
    val_start_idx = N_train_max + N_test
    for (
        reduced_dims_tuple_i,
        reduced_data_i,
    ) in reduced_all_data.items():
        reduced_train_test_val[reduced_dims_tuple_i] = {
            "train": dict(),
            "test": dict(),
            "val": dict(),
        }

        for data_key in ["encoded_Jacobians", "encoded_inputs", "encoded_outputs", 'inputs', 'outputs']:
            data_values = reduced_data_i.get(data_key)
            if data_values is not None:
                reduced_train_test_val[reduced_dims_tuple_i]["train"][data_key] = data_values[:N_train_max]
                reduced_train_test_val[reduced_dims_tuple_i]["test"][data_key] = data_values[N_train_max:val_start_idx]
                reduced_train_test_val[reduced_dims_tuple_i]["val"][data_key] = data_values[
                    val_start_idx : val_start_idx + N_val
                ]

                if save_path:
                    Path(save_path, dimension_reduction_type, "training").mkdir(parents=True, exist_ok=True)
                    Path(save_path, dimension_reduction_type, "testing").mkdir(parents=True, exist_ok=True)
                    Path(save_path, dimension_reduction_type, "validation").mkdir(parents=True, exist_ok=True)

                    # saves different sets of training data for easy loading from file
                    for N_train in N_trains:
                        np.save(
                            Path(
                                save_path,
                                dimension_reduction_type,
                                "training",
                                f"{reduced_dims_tuple_i}_{data_key}_{N_train}.npy",
                            ),
                            data_values[0:N_train],
                        )
                    np.save(
                        Path(save_path, dimension_reduction_type, "testing", f"{reduced_dims_tuple_i}_{data_key}.npy"),
                        data_values[N_train_max:val_start_idx],
                    )
                    np.save(
                        Path(
                            save_path, dimension_reduction_type, "validation", f"{reduced_dims_tuple_i}_{data_key}.npy"
                        ),
                        data_values[val_start_idx : val_start_idx + N_val],
                    )

    return reduced_train_test_val  # keeps all N_train_max data in the training data set



from pathlib import Path
from typing import Dict, Iterable, Tuple

import hickle
import jax.random as jr
from jax import default_device as jax_default_device
from jax import device_get, device_put
from jax import devices as jax_devices
from jax.tree_util import tree_map
from jax.typing import ArrayLike as JAXArrayLike
from jaxtyping import PyTree
from numpy.typing import ArrayLike


def dump_pytree_to_disk(pytree: PyTree, path: Path) -> None:
    import numpy as np

    hickle.dump(tree_map(np.asarray, pytree), path, mode="w")


def load_pytree_from_disk(path: Path) -> PyTree:
    import jax.numpy as jnp

    return tree_map(jnp.array, hickle.load(path))


def load_pytree_as_dict_of_np_arrays_from_disk(
    path: Path,
) -> Dict[str, ArrayLike]:
    return hickle.load(path)


def device_put_pytree(pytree: PyTree) -> PyTree:
    return tree_map(device_put, pytree)


def device_get_pytree(pytree: PyTree) -> PyTree:
    return tree_map(device_get, pytree)


def add_squared_norms_of_each_entry(
    data_dict: Dict[str, JAXArrayLike],
) -> Dict[str, JAXArrayLike]:
    device = data_dict[list(data_dict.keys())[0]].device()
    if device in jax_devices("cpu"):
        from .losses import __cpu_vectorized_squared_norm

        norm_func = __cpu_vectorized_squared_norm
    else:
        from .losses import __vectorized_squared_norm

        norm_func = __vectorized_squared_norm
    for data_key in list(data_dict.keys()):
        if data_key != "X":
            data_dict[f"{data_key}_norms"] = norm_func(data_dict[data_key])
    return data_dict


def load_encoded_training_validation_and_testing_data(
    data_path: Path,
    REDUCED_DIMS: Tuple[int, int],
    training_data_keys: Iterable[str],
    N_TRAIN: int,
    renaming_dict: Dict[str, str] = None,
    test_data_keys: Iterable[str] = ("encoded_inputs", "encoded_outputs", "encoded_Jacobians"),
    N_VAL: int = 2500,
    N_TEST: int = 10_000,  # 25_000,
) -> Dict[str, Dict[str, ArrayLike]]:
    import jax.numpy as jnp

    """
    Loads training and testing data from .npy files and computes squared L2 norms for testing data.

    Parameters:
        REDUCED_DIMS (Tuple[int, int]): Reduced dimensions.
        data_keys (Iterable[str]): Data keys to load.
        N_TRAIN (int): Number of training samples.
        N_TEST (int, optional): Number of testing samples (default is 25_000; truncated if lower).

    Returns:
        Dict[str, Dict[str, np.ndarray]]:
            - "training data": Maps each key to its training array.
            - "testing data": Maps each key to its testing array and includes computed norms (keyed as "{data_key}_norms").
    """
    if renaming_dict is None:
        renaming_dict = {data_key: data_key for data_key in training_data_keys}

    # Load training data onto GPU-- useful only if training data saved to "training/f"{REDUCED_DIMS}_{data_key}_{N_TRAIN}.npy"
    train_data = {
        renaming_dict[data_key]: jnp.load(Path(data_path, "training", f"{REDUCED_DIMS}_{data_key}_{N_TRAIN}.npy"))
        for data_key in training_data_keys
    }

    with jax_default_device(jax_devices("cpu")[0]):
        cpu_train_data = {
            renaming_dict[data_key]: jnp.load(Path(data_path, "training", f"{REDUCED_DIMS}_{data_key}_{N_TRAIN}.npy"))
            for data_key in training_data_keys
        }

    # print("Taking stock of training data:")
    # for key, val in train_data.items():
    #     print(f"\t{key}, shape: {val.shape}")
    # Load validation data onto GPU; truncate if N_VAL is less than 2500
    if N_VAL > 0:
        if N_VAL < 2500:
            val_data = {
                renaming_dict[data_key]: jnp.load(Path(data_path, "validation", f"{REDUCED_DIMS}_{data_key}.npy"))[
                    :N_VAL
                ]
                for data_key in training_data_keys
            }
        else:
            val_data = {
                renaming_dict[data_key]: jnp.load(Path(data_path, "validation", f"{REDUCED_DIMS}_{data_key}.npy"))
                for data_key in training_data_keys
            }

    # Load testing data onto CPU; truncate if N_TEST is less than 25000
    # by default includes encoded_Jacobians, since we want to test the accuracy of Jacobians even when training without them
    with jax_default_device(jax_devices("cpu")[0]):
        if N_TEST < 10_000:
            cpu_test_data = {
                renaming_dict[data_key]: jnp.load(Path(data_path, "testing", f"{REDUCED_DIMS}_{data_key}.npy"))[:N_TEST]
                for data_key in test_data_keys
            }
        else:
            cpu_test_data = {
                renaming_dict[data_key]: jnp.load(Path(data_path, "testing", f"{REDUCED_DIMS}_{data_key}.npy"))
                for data_key in test_data_keys
            }

    # Compute squared L2 or Frobenius norms for each validation/testing data array and add to dictionaries.
    train_val_test = {
        "train": train_data,
        "train_cpu": add_squared_norms_of_each_entry(cpu_train_data),
        "test": add_squared_norms_of_each_entry(cpu_test_data),
    }
    if N_VAL > 0:
        train_val_test["val"] = add_squared_norms_of_each_entry(val_data)
    return train_val_test


def load_encoded_data_and_add_noise(
    data_path: Path, reduced_dims: Tuple[int, int], n_train: int, observation_noise_key: jr.PRNGKey
):
    training_data_keys = ("encoded_inputs", "encoded_outputs")
    test_data_keys = ("encoded_inputs", "encoded_outputs", "encoded_Jacobians")
    renaming_dict = {"encoded_inputs": "X", "encoded_outputs": "fX", "encoded_Jacobians": "dfXdX"}
    train_val_test = load_encoded_training_validation_and_testing_data(
        data_path=data_path,
        REDUCED_DIMS=reduced_dims,
        training_data_keys=training_data_keys,
        test_data_keys=test_data_keys,
        renaming_dict=renaming_dict,
        N_TRAIN=n_train,
    )
    TRAINING_DATA, VALIDATION_DATA, CPU_TEST_DATA = (
        train_val_test["train"],
        train_val_test["val"],
        train_val_test["test"],
    )

    # Vanilla amortized estimation trains with non-degenerate X,Y distribution data.
    # In our case, we need to add noise from our noise model: Y = f(X) + n, n\sim N(0, C).
    # Since we actually train on !encoded! data, with whitening encoding transformation y_r = E Y = Ef(x) + En, where E D = I_r, D = C E^T ,
    # we arrive at y_r \sim N(Ef(x), E C E.T) = N(Ef(x), I_r). Hence, we add white noise
    key_train, key_val = jr.split(observation_noise_key)
    TRAINING_DATA["Y"] = TRAINING_DATA["fX"] + jr.normal(key_train, TRAINING_DATA["fX"].shape)
    VALIDATION_DATA["Y"] = VALIDATION_DATA["fX"] + jr.normal(key_val, VALIDATION_DATA["fX"].shape)
    del TRAINING_DATA["fX"], VALIDATION_DATA["fX"]

    return TRAINING_DATA, VALIDATION_DATA, CPU_TEST_DATA, renaming_dict


def check_batch_size_validity(*, data_iterable: Iterable[ArrayLike], batch_size: int) -> int:
    n_batches_list = []
    n_data = next(iter(data_iterable)).shape[0]
    for data_tuple in data_iterable:
        n_data_i = data_tuple.shape[0]
        assert n_data_i == n_data, "We require the same number of data for inputs and all outputs"
        n_batches, remainder = divmod(n_data_i, batch_size)
        if remainder != 0:
            raise ValueError("Adjust `batch_size` to evenly divide training and testing data.")
        n_batches_list.append(n_batches)
    assert n_batches_list[:-1] == n_batches_list[1:], "The data does not all have the same number of elements!"
    return n_batches



from typing import Callable

import jax

jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from equinox import filter_grad, filter_jit, filter_vmap
from jax import jit, vjp

from .equinox_nn_factories import eqxModule

__squared_norm = lambda x: jnp.sum(x**2)
__vectorized_squared_norm = jax.vmap(jit(__squared_norm, device=jax.devices("gpu")[0]), in_axes=0)
__cpu_vectorized_squared_norm = jax.vmap(jit(__squared_norm, device=jax.devices("cpu")[0]), in_axes=0)
__divide = jit(lambda scores, normalizations: scores / normalizations, device=jax.devices("gpu")[0])
__cpu_divide = jit(
    lambda scores, normalizations: scores / normalizations,
    device=jax.devices("cpu")[0],
)
# This is only an Monte Carlo estimate of the L2 Bochner loss, i.e. just an empirical MSE.
L2_Bochner_loss: Callable[[eqxModule, jax.Array, jax.Array], float] = filter_jit(
    lambda nn, input_X, actual_Y: jnp.mean(jnp.sum((filter_vmap(nn)(input_X) - actual_Y) ** 2, axis=1))
)
vectorized_grad_L2_Bochner_loss = filter_grad(L2_Bochner_loss)


@filter_jit(backend="gpu")
def vectorized_H1_Bochner_loss(
    nn: eqxModule,
    input_X: jax.Array,
    actual_Y: jax.Array,
    actual_dYdX: jax.Array,
) -> float:
    predicted_Y, predicted_dYdX = __vectorized_value_and_jacrev(nn, input_X)
    return jnp.mean(
        __vectorized_squared_norm(actual_Y - predicted_Y) + __vectorized_squared_norm(actual_dYdX - predicted_dYdX)
    )


vectorized_grad_H1_Bochner_loss = filter_grad(vectorized_H1_Bochner_loss)


@filter_jit(backend="gpu")
def __vectorized_value_and_jacrev(f: Callable[[jax.Array], jax.Array], xs: jax.Array) -> jax.Array:
    """
    Compute function values and full Jacobians for a batch of inputs using vjp.

    Parameters:
        f (Callable[[jax.Array], jax.Array]): Function to evaluate.
        xs (jax.Array): Batch of inputs.

    Returns:
        jax.Array: Array of tuples (f(x), Jacobian at x).

    Notes:
        Uses JAX's vjp with an identity matrix and vmap for efficient batch processing.
    """
    _, pullback = vjp(f, xs[0])
    basis = jnp.eye(_.size, dtype=_.dtype)  # Create an identity matrix the size of the output

    def value_and_jacrev_x(x: jax.Array):
        """
        Helper function to compute function value and Jacobian at a given input.
        """
        y, pullback = vjp(f, x)
        return (
            y,
            filter_vmap(pullback)(basis)[0],
        )  # Extract the single Jacobian matrix from the vmap result

    return filter_vmap(value_and_jacrev_x)(xs)


@filter_jit(backend="cpu")
def __cpu_vectorized_value_and_jacrev(f: Callable[[jax.Array], jax.Array], xs: jax.Array) -> jax.Array:
    """
    Compute function values and full Jacobians for a batch of inputs using vjp.

    Parameters:
        f (Callable[[jax.Array], jax.Array]): Function to evaluate.
        xs (jax.Array): Batch of inputs.

    Returns:
        jax.Array: Array of tuples (f(x), Jacobian at x).

    Notes:
        Uses JAX's vjp with an identity matrix and vmap for efficient batch processing.
    """
    _, pullback = vjp(f, xs[0])
    basis = jnp.eye(_.size, dtype=_.dtype)  # Create an identity matrix the size of the output

    def value_and_jacrev_x(x: jax.Array):
        """
        Helper function to compute function value and Jacobian at a given input.
        """
        y, pullback = vjp(f, x)
        return (
            y,
            filter_vmap(pullback)(basis)[0],
        )  # Extract the single Jacobian matrix from the vmap result

    return filter_vmap(value_and_jacrev_x)(xs)


@filter_jit(backend="cpu")
def cpu_compute_H1_bochner_relative_errors(
    nn: eqxModule,
    X: jax.Array,
    fX: jax.Array,
    dfXdX: jax.Array,
    fX_L2_norms: jax.Array,
    dfXdX_Frobenius_norms: jax.Array,
) -> jax.Array:
    """
    Compute RMSE for outputs and Jacobians, plus their normalized errors.

    Parameters:
        nn: eqxModule - Model predicting outputs and Jacobians.
        X: jax.Array - Inputs.
        fX: jax.Array - Target outputs.
        dfXdX: jax.Array - Target Jacobians.
        fX_L2_norms: jax.Array - L2 norms of outputs.
        dfXdX_Frobenius_norms: jax.Array - Frobenius norms of Jacobians.

    Returns:
        jax.Array: Tuple (RMSE(NN(X);fX), RMSE(JacNN(X); dfXdX), normalized RMSE(NN), normalized RMSE(JacNN).
    """

    predicted_fX, predicted_dfXdX = __cpu_vectorized_value_and_jacrev(nn, X)
    Yhat_squared_norm_errors = __cpu_vectorized_squared_norm(predicted_fX - fX)
    dYhatdX_squared_norm_errors = __cpu_vectorized_squared_norm(predicted_dfXdX - dfXdX)

    return (
        jnp.sqrt(jnp.mean(Yhat_squared_norm_errors)),  #  Root mean (function) squared error
        jnp.sqrt(jnp.mean(dYhatdX_squared_norm_errors)),  # Root mean (Jacobian) squared error
        # Root mean relative (function) squared error
        jnp.sqrt(jnp.mean(__cpu_divide(Yhat_squared_norm_errors, fX_L2_norms))),
        # Root mean relative (Jacobian) squared error
        jnp.sqrt(jnp.mean(__cpu_divide(dYhatdX_squared_norm_errors, dfXdX_Frobenius_norms))),
    )


@filter_jit(backend="cpu")
def cpu_compute_L2_bochner_relative_errors(
    nn: eqxModule,
    X: jax.Array,
    fX: jax.Array,
    fX_L2_norms: jax.Array,
) -> jax.Array:
    """
    Compute RMSE for outputs and Jacobians, plus their normalized errors.

    Parameters:
        nn: eqxModule - Model predicting outputs and Jacobians.
        X: jax.Array - Inputs.
        fX: jax.Array - Target outputs.
        dfXdX: jax.Array - Target Jacobians.
        fX_L2_norms: jax.Array - L2 norms of outputs.
        dfXdX_Frobenius_norms: jax.Array - Frobenius norms of Jacobians.

    Returns:
        jax.Array: Tuple (RMSE(NN(X);fX), RMSE(JacNN(X); dfXdX), normalized RMSE(NN) normalized RMSE(JacNN).
    """
    Yhat_squared_norm_errors = __cpu_vectorized_squared_norm(filter_vmap(nn)(X) - fX)
    return (
        jnp.sqrt(jnp.mean(Yhat_squared_norm_errors)),  #  Root mean (function) squared error
        # Root mean relative (function) squared error
        jnp.sqrt(jnp.mean(__cpu_divide(Yhat_squared_norm_errors, fX_L2_norms))),
        # Root mean relative Jacobian squared error
    )


@filter_jit(backend="gpu")
def compute_H1_bochner_relative_errors(
    nn: eqxModule,
    X: jax.Array,
    fX: jax.Array,
    dfXdX: jax.Array,
    fX_L2_norms: jax.Array,
    dfXdX_Frobenius_norms: jax.Array,
) -> jax.Array:
    """
    Compute RMSE for outputs and Jacobians, plus their normalized errors.

    Parameters:
        nn: eqxModule - Model predicting outputs and Jacobians.
        X: jax.Array - Inputs.
        fX: jax.Array - Target outputs.
        dfXdX: jax.Array - Target Jacobians.
        fX_L2_norms: jax.Array - L2 norms of outputs.
        dfXdX_Frobenius_norms: jax.Array - Frobenius norms of Jacobians.

    Returns:
        jax.Array: Tuple (RMSE(NN(X);fX), RMSE(JacNN(X); dfXdX), normalized RMSE(NN) normalized RMSE(JacNN).
    """

    predicted_fX, predicted_dfXdX = __vectorized_value_and_jacrev(nn, X)
    Yhat_squared_norm_errors = __vectorized_squared_norm(predicted_fX - fX)
    dYhatdX_squared_norm_errors = __vectorized_squared_norm(predicted_dfXdX - dfXdX)

    return (
        jnp.sqrt(jnp.mean(Yhat_squared_norm_errors)),  #  Root mean (function) squared error
        jnp.sqrt(jnp.mean(dYhatdX_squared_norm_errors)),  # Root mean (Jacobian) squared error
        # Root mean relative (function) squared error
        jnp.sqrt(jnp.mean(__divide(Yhat_squared_norm_errors, fX_L2_norms))),
        # Root mean relative Jacobian squared error
        jnp.sqrt(jnp.mean(__divide(dYhatdX_squared_norm_errors, dfXdX_Frobenius_norms))),
    )


@filter_jit(backend="gpu")
def compute_L2_bochner_relative_errors(
    nn: eqxModule,
    X: jax.Array,
    fX: jax.Array,
    fX_L2_norms: jax.Array,
) -> jax.Array:
    """
    Compute RMSE for outputs and Jacobians, plus their normalized errors.

    Parameters:
        nn: eqxModule - Model predicting outputs and Jacobians.
        X: jax.Array - Inputs.
        fX: jax.Array - Target outputs.
        dfXdX: jax.Array - Target Jacobians.
        fX_L2_norms: jax.Array - L2 norms of outputs.
        dfXdX_Frobenius_norms: jax.Array - Frobenius norms of Jacobians.

    Returns:
        jax.Array: Tuple (RMSE(NN(X);fX), RMSE(JacNN(X); dfXdX), normalized RMSE(NN) normalized RMSE(JacNN).
    """
    Yhat_squared_norm_errors = __vectorized_squared_norm(filter_vmap(nn)(X) - fX)
    return (
        jnp.sqrt(jnp.mean(Yhat_squared_norm_errors)),  #  Root mean (function) squared error
        # Root mean relative (function) squared error
        jnp.sqrt(jnp.mean(__divide(Yhat_squared_norm_errors, fX_L2_norms))),
        # Root mean relative Jacobian squared error
    )
